
require "walk"

module CodePreloader
  # Manage a list of files
  class FileList 
    alias Filter = String -> Bool

    class NotADirectory < Exception
      def initialize(path)
        super(path.to_s)
      end
    end

    @sources : Array(String)
    @include_filters : Array(Filter)
    @exclude_filters : Array(Filter)

    def initialize(dirs = [] of String)
      @sources = [] of String
      @include_filters = [] of Filter
      @exclude_filters = [] of Filter
      dirs.each { |dir| self.add(dir) }
    end

    def add(dirs : Array(String))
      dirs.each { |dir| add(dir) }
    end

    def add(dir : String)
      raise NotADirectory.new(dir) if !File.exists? dir

      @sources << dir
    end

    def select(&filter : Filter)
      @include_filters << filter
    end

    def reject(&filter : Filter)
      @exclude_filters << filter
    end

    def each(&block)
      # ensure we display files only once
      seen = Set(String).new

      # walk each source
      @sources.sort.each do |dir|
        walker = Walk::Down.new(dir)

        walker = walker.filter do |path|
          is_dir = File.directory? path
          keep = true
          must_select = false
          must_reject = false
          clean_path = path.to_s.gsub(/^\.\//,"")

          @include_filters.each do |filter_in|
            must_select = must_select || filter_in.call(clean_path)
          end
          keep = keep && must_select if @include_filters.any?
          keep = keep || is_dir

          @exclude_filters.each do |filter_out|
            must_reject = must_reject || filter_out.call(clean_path)
          end
          keep = keep && !must_reject if @exclude_filters.any?

          keep
        end

        walker.each do |path|
          clean_path = path.to_s.gsub(/^\.\//,"")
          next if File.directory? clean_path

          path = File.realpath(path) if File.symlink? clean_path
          next if seen.includes? clean_path

          seen << clean_path
          yield clean_path
        end
      end
    end

    def to_a()
      files = [] of String
      self.each do |path|
        files << path.to_s
      end
      files.sort
    end
  end
end
